/*
  wavelet.cpp

  discrete wavlet transformation (DWT)

  Copyright (C) 2004, Patrick Rammelt <rammelt@cs.tu-berlin.de>
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
  wavelet.cpp

  Discrete Wavelet Transformations

  Patrick Rammelt - 25.03.2004
*/

#ifndef _WAVELET_CPP
#define _WAVELET_CPP

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "mathematical.cpp"


/* =========================================================================
   Discrete Wavelet Transformation - Filter-Coefficients for special filters
   ========================================================================= */
/* -------------------------------------------------------------------------
   Debauchy 9/7
   ------------------------------------------------------------------------- */
extern const long   dwt_Debauchy97_size[2] = { 9, 7 };
extern const double dwt_Debauchy97_low[9]  = { 0.02674875741080976,
					       -0.01686411844287495,
					       -0.07822326652898785,
					       0.2668641184428723,
					       0.6029490182363579,
					       0.2668641184428723,
					       -0.07822326652898785,
					       -0.01686411844287495,
					       0.02674875741080976 };
extern const double dwt_Debauchy97_high[7] = { 0.09127176311424948,
					       -0.05754352622849957,
					       -0.5912717631142470,
					       1.115087052456994,
					       -0.5912717631142470,
					       -0.05754352622849957,
					       0.09127176311424948 };

/* -------------------------------------------------------------------------
   lifting scheme filter 5/3
   ------------------------------------------------------------------------- */
extern const long   dwt_filter53_size[2] = { 5, 3 };
extern const double dwt_filter53_low[5]  = { -1.0/8.0,
					     2.0/8.0,
					     6.0/8.0,
					     2.0/8.0,
					     -1.0/8.0 };
extern const double dwt_filter53_high[3] = { -1.0/2.0, 
					     1.0, 
					     -1.0/2.0 }; 


/* =========================================================================
   Discrete Wavelet Transformation - Filter
   ( base class )
   ========================================================================= */
class dwtFilter {

protected:

  /* -----------------------------------------------------------------------
     attributes
     ----------------------------------------------------------------------- */
  double *c[2]; // filter-coefficients ( low-pass / high-pass )
  long    n[2]; // number of filter-coefficients

public:

  /* -----------------------------------------------------------------------
     construct / destruct
     ----------------------------------------------------------------------- */
  dwtFilter (void) { c[0] = c[1] = NULL; n[0] = n[1] = 0; }
  virtual ~dwtFilter (void) { unset(); }

  /* -----------------------------------------------------------------------
     main set method
       low       : low-pass-filter-coefficients
       lowOrder  : number of coefficients in low
       high      : high-pass-filter-coefficients
       highOrder : number of coefficients in high
     ----------------------------------------------------------------------- */
  virtual void set (const double *low,  const long lowOrder,
		    const double *high, const long highOrder);

  /* -----------------------------------------------------------------------
     set / unset
     ----------------------------------------------------------------------- */
  virtual void set (const double *low, const double *high, const long size[2]){
    set (low, size[0], high, size[1]); }
  virtual void set (const class dwtFilter &flt) {
    set (flt.c[0], flt.n[0], flt.c[1], flt.n[1]); }
  virtual void unset (void);

  /* -----------------------------------------------------------------------
     interface to the filter-method of the derived class ( prototype )
       sig[0] : un-filtered signal (input)
       sig[1] : filtered signal (output)
       len    : length of sig[0], length of sig[1] = len/2
       step   : sample distance in signals
       fi     : filter-selection ( low-pass = 0 / high-pass-filter = 1 )
       offset : index of first sample in signals
     ----------------------------------------------------------------------- */
  virtual inline void apply (double *sig[], const long len, const long step, 
			     const int fi, const long offset=0) const {}

  /* -----------------------------------------------------------------------
     filtering  - single level ( prototype )
       sig    : signals
       len    : length of signals
       step   : sample distance in signals
       offset : index of first sample in signals
       mode   : ( used in different ways )
     ----------------------------------------------------------------------- */
  virtual inline void single (double *sig[], const long len, 
			      const long step=1, const long offset=0, 
			      const unsigned int mode=0) const {}

  /* -----------------------------------------------------------------------
     filtering - multi level ( prototype )
       sig    : signals
       size   : length of signal sig[0] (in each dimension)
                i.e. sig[0] is a vector, square, cube or hypercube
       dims   : number of dimensions of signal sig[0]
       depth  : number of steps per dim. 
                (depth<=0 : max possible number + depth)
       return : pointer to result (current sig[0])
     ----------------------------------------------------------------------- */
 virtual double* multi (double *sig[], const long size, 
			const long dims=1, long depth=0) const { 
   return (NULL); }

  /* -----------------------------------------------------------------------
     max depth = max possible len/=2 operations 
     (where len%2==0 - i.e. len must be even)
     ----------------------------------------------------------------------- */
  virtual long maxDepth (unsigned long len) const {
    for (long i=0; len; len=len>>1, i++) if (len&(unsigned long)1) return (i); 
    return (0); }

  /* -----------------------------------------------------------------------
     get depth for given length len and depth
     ----------------------------------------------------------------------- */
  virtual long getDepth (unsigned long len, long depth=0) const {
    const long maximum = maxDepth(len);
    if (depth<=0) return (maximum+depth);
    else if (depth>maximum) return(maximum); 
    else return (depth); }

protected:

  /* -----------------------------------------------------------------------
     (re-)initialize ( allocate memory )
     ----------------------------------------------------------------------- */
  virtual void create (const long lowOrder, const long highOrder);

  /* -----------------------------------------------------------------------
     revert filter-coefficients 
     ----------------------------------------------------------------------- */
  inline void revert (double *coeff, const long order);

  /* -----------------------------------------------------------------------
     init parameters for recursive tile(), init() and copy()
       len[d]         : length of dimension d (output)
       step[d]        : sample distance in dimension d (output)
       dims           : number of dimensions
       virtualLength  : virtual (current) length of each dimension
       physicalLength : physical (max) length of each dimension
     ----------------------------------------------------------------------- */
  virtual void initSize (long *len, long *step, long dims, 
			 long virtualLength, long physicalLength) const;

  /* -----------------------------------------------------------------------
     tile dimension tiledim by applying single()
       sig     : signals for single() method
       mode    : ( used in different ways by specialized single()-methods )
       dim     : current dimension-index (init-value=highest dim)
       tiledim : index of dimension to be (un-)tiled
       offset  : offset of first element
       len[d]  : length of dimension d
       step[d] : sample distance in dimension d
     ----------------------------------------------------------------------- */
  virtual void tile (double *sig[], const unsigned int mode, 
		     const long dim, const long tiledim, 
		     long offset, const long *len, const long *step) const;

  /* -----------------------------------------------------------------------
     shift
       sig[]     : signals
       sft[]     : signals to shift (sig[sft[0..m-1]])
       m         : number of signals to shift
       levels    : number of levels
       start     : first level to work on
       dims      : number of dimensions
       len[d]    : length of dimension d
       step[d]   : sample distance in dimension d
       mode      : mode for tile()
       return    : last tiled dimension
     ----------------------------------------------------------------------- */
  inline long shift (double *sig[], const int sft[], const int m,
		     const long levels, const long start,
		     const long dims, long len[], const long step[], 
		     const unsigned int mode) const;

  /* -----------------------------------------------------------------------
     offset for shifting
       d         : current dimension
       dims      : number of dimensions
       len[d]    : length of dimension d
       step[d]   : sample distance in dimension d
     ----------------------------------------------------------------------- */
  inline long shiftOffset (const long d, const long dims,
			   const long len[], const long step[]) const {
    const long prev = index(d-1,dims); return (len[prev] * step[prev]); }

  /* -----------------------------------------------------------------------
     init all x[i] = v
       x       : signal (output)
       v       : init value
       dim     : current dimension-index (init-value=highest dim)
       offset  : offset of first element
       len[d]  : length of dimension d
       step[d] : sample distance in dimension d
     ----------------------------------------------------------------------- */
  void init (double *x, double v, const long dim, 
	     long offset, const long *len, const long *step) const;

  /* -----------------------------------------------------------------------
     init signal ( set all elements = 0 )
       x    : signal
       v    : value
       len  : length of x
       step : sample distance in x
     ----------------------------------------------------------------------- */
  virtual inline void init (double *x, double v, 
			    const long len, const long step) const {
    for (long i=0; i<len; x[i*step] = v, i++); }

  /* -----------------------------------------------------------------------
     copy x -> y
       x       : signal (input)
       y       : signal (output)
       dim     : current dimension-index (init-value=highest dim)
       offset  : offset of first element
       len[d]  : length of dimension d
       step[d] : sample distance in dimension d
     ----------------------------------------------------------------------- */
  void copy (double *x, double *y, const long dim, 
	     long offset, const long *len, const long *step) const;

  /* -----------------------------------------------------------------------
     copy x -> y  ( y[i*step] = x[i*step], i=0...len-1 )
       x    : signal (input)
       y    : signal (output)
       len  : number of elements to be copied
       step : sample distance in x and y 
     ----------------------------------------------------------------------- */
  inline void copy (double *x,double *y, const long len,const long step) const{
    for (long i=0; i<len; y[i*step] = x[i*step], i++); }

  /* -----------------------------------------------------------------------
     2^p
     ----------------------------------------------------------------------- */
  inline unsigned long power2 (const unsigned long p) const { 
    return ((long)1<<p); }

  /* -----------------------------------------------------------------------
     get index "go round" ( change j to a position inside a vector v[0...m-1] )
       j : element number ( might be negative or j>=m )
           e.g. -1 -> m-1 , -2 -> m-2 , m -> 0 , m+1 -> 1
       m : number of elements in vector v
     ----------------------------------------------------------------------- */
  inline unsigned long index (const long j, const long m) const { 
    long i = j%m;                           // prepare for different behaviors
    return ((i<0 ? i+m : i)); }             // for j % m , when j<0

  /* -----------------------------------------------------------------------
     even / odd 
     ----------------------------------------------------------------------- */
  inline bool even (unsigned long x) const { return (!(x&(unsigned long)1)); }
  inline bool odd  (unsigned long x) const { return ( (x&(unsigned long)1)); }

  /* -----------------------------------------------------------------------
     median
     ----------------------------------------------------------------------- */
  inline double median (const double a, const double b, const double c) const {
    if (a>b) {
      if (b>c) return (b); else if (c>a) return (a); else return (c);
    } else {
      if (c>b) return (b); else if (a>c) return (a); else return (c);
    }
  }

  /* -----------------------------------------------------------------------
     filter offset = half high-pass-filter-order
     ----------------------------------------------------------------------- */
  virtual inline long filterOffset (void) const { return (n[1]/2); }

  /* -----------------------------------------------------------------------
     signal offset for low-pass-signal (fi==0) / high-pass-signal (fi==1)
     ----------------------------------------------------------------------- */
  virtual inline long signalOffset (const long len, const int fi) const { 
    return ((fi ? len/2 : 0)); }

  /* -----------------------------------------------------------------------
     toggle values a, b
     ----------------------------------------------------------------------- */
  inline void toggle (double &a, double &b) const { double c=a; a=b; b=c; }
  inline void toggle (long &a, long &b) const { long c=a; a=b; b=c; }

  /* -----------------------------------------------------------------------
     toggle signals sig[a], sig[b]
     ----------------------------------------------------------------------- */
  inline void toggleSignals (double *sig[], long a=0, long b=1) const { 
    double *s = sig[a]; sig[a] = sig[b]; sig[b] = s; }
};


/* -------------------------------------------------------------------------
   set
   ------------------------------------------------------------------------- */
void dwtFilter::set (const double *low,  const long lowOrder,
		     const double *high, const long highOrder) 
{
  create (lowOrder,highOrder); 
  for (long i=0; i<n[0]; c[0][i]=low[i],  i++); 
  for (long i=0; i<n[1]; c[1][i]=high[i], i++); 
}


/* -------------------------------------------------------------------------
   (re-)initialize ( allocate memory )
   ------------------------------------------------------------------------- */
void dwtFilter::create (const long lowOrder, const long highOrder) { 
  unset(); 
  n[0] = lowOrder; 
  n[1] = highOrder; 
  c[0] = new double[n[0]]; 
  c[1] = new double[n[1]]; 
}


/* -------------------------------------------------------------------------
   revert filter-coefficients 
   ------------------------------------------------------------------------- */
inline void dwtFilter::revert (double *coeff, const long order) 
{
  for (long i=1, n2=(order-1)/2; i<=n2; i++) {
    double signum = (i&1 ? -1 : 1);              // odd => -1, even => 1
    coeff[n2+i] *= signum;
    coeff[n2-i] *= signum;
    toggle (coeff[n2+i],coeff[n2-i]);
  }
}


/* -------------------------------------------------------------------------
   unset ( free memory )
   ------------------------------------------------------------------------- */
void dwtFilter::unset (void) { 
  if (c[0]!=NULL) delete[] c[0];
  if (c[1]!=NULL) delete[] c[1];
  c[0] = c[1] = NULL; n[0] = n[1] = 0; 
}


/* -------------------------------------------------------------------------
   init parameters for recursive tile()
   ------------------------------------------------------------------------- */
void dwtFilter::initSize (long *len, long *step, long dims, 
			  long virtualLength, long physicalLength) const 
{
  len[0]  = virtualLength;
  step[0] = 1;
  for (long d=1; d<dims; d++) { 
    step[d] = physicalLength * step[d-1];
    len[d]  = virtualLength; 
  }
}


/* -------------------------------------------------------------------------
   tile dimension tiledim by applying single()
   ------------------------------------------------------------------------- */
void dwtFilter::tile (double *sig[], const unsigned int mode, 
		      const long dim, const long tiledim, 
		      long offset, const long *len, const long *step) const
{
  if (dim<0) {

    // apply step ( call specialized single-step-method for tiling )
    single (sig, len[tiledim], step[tiledim], offset, mode);

  } else if (dim==tiledim) {

    // do not step into tiledim - go on with next dimension
    tile (sig, mode, dim-1, tiledim, offset, len, step);

  } else {

    // tile all states of dimension dim
    for (long i=0; i<len[dim]; offset+=step[dim], i++) {
      tile (sig, mode, dim-1, tiledim, offset, len, step);
    }

  }
}


/* -------------------------------------------------------------------------
   shift
   ------------------------------------------------------------------------- */
inline long dwtFilter::shift (double *sig[], const int sft[], const int m,
			      const long levels, const long start, 
			      const long dims, long len[], const long step[], 
			      const unsigned int mode) const
{
  // rem first elements of signals (->offsetting)
  double *sig0[m];
  for (long i=0; i<m; sig0[i] = sig[sft[i]], i++);

  // do levels of shifting
  long d = -1;
  for (long lev=0; lev<levels; lev++) {
    d = lev % dims;
    if (lev>=start) {
      const long offset = shiftOffset (d,dims,len,step);
      for (long i=0; i<m; sig[sft[i]] = &(sig0[i][offset]), i++);
      tile (sig, mode, dims-1, d, 0, len, step);
    }
    len[d] /= 2;
  }

  // undo offsetting & return last tiled dimension   
  for (long i=0; i<m; sig[sft[i]] = sig0[i], i++);
  return (d);
}


/* -------------------------------------------------------------------------
   init all x[i] = v
   ------------------------------------------------------------------------- */
void dwtFilter::init (double *x, double v, const long dim, 
		      long offset, const long *len, const long *step) const 
{
  if (dim<1) {
    init (&(x[offset]),v,len[0],step[0]);    // simple init in first dimension
  } else {
    for (long i=0; i<len[dim]; offset+=step[dim], i++) {
      init (x, v, dim-1, offset, len, step);    // all states of dimension dim
    }
  }
}


/* -------------------------------------------------------------------------
   copy x -> y
  ------------------------------------------------------------------------- */
void dwtFilter::copy (double *x, double *y, const long dim, 
		      long offset, const long *len, const long *step) const 
{
  if (dim<1) {
    copy (&(x[offset]),&(y[offset]),len[0],step[0]); // simple copy in 1st dim.
  } else {
    for (long i=0; i<len[dim]; offset+=step[dim], i++) {
      copy (x, y, dim-1, offset, len, step);    // all states of dimension dim
    }
  }
}


/* =========================================================================
   Discrete Wavelet Transformation - Filter
   ( analysis filter )
   ========================================================================= */
class dwtAnalyzer : public dwtFilter {

public:

  /* -----------------------------------------------------------------------
     construct / destruct
       low       : low-pass-filter-coefficients 
       lowOrder  : number of coefficients in low
       high      : high-pass-filter-coefficients
       highOrder : number of coefficients in high
     ----------------------------------------------------------------------- */
  dwtAnalyzer (void) {}
  dwtAnalyzer (const class dwtAnalyzer &flt) { set (flt); }
  dwtAnalyzer (const double *low,  const long lowOrder,
	       const double *high, const long highOrder) { 
    set (low,lowOrder, high,highOrder); }
  dwtAnalyzer (const double *low, const double *high, const long size[2]) {
    set (low, size[0], high, size[1]); }

  /* -----------------------------------------------------------------------
     interface to the filter
       sig[0] : un-filtered signal (input)
       sig[1] : filtered signal (output)
       len    : length of sig[0], length of sig[1] = len/2
       step   : sample distance in signals
       fi     : filter-selection ( low-pass = 0 / high-pass-filter = 1 )
       offset : index of first sample in signals
     ----------------------------------------------------------------------- */
  virtual inline void apply (double *sig[], const long len, const long step, 
			     const int fi, const long offset=0) const {
    analyze (sig, len, step, fi, offset, 
	     filterOffset(), signalOffset(len,fi)*step); }

  /* -----------------------------------------------------------------------
     analyze filter
       sig[0] : un-filtered signal (input)
       sig[1] : filtered signal (output)
       len    : length of sig[0], length of sig[1] = len/2
       step   : sample distance in signals
       fi     : filter-selection ( low-pass = 0 / high-pass-filter = 1 )
       offset : index of first sample in signals
       fd     : filter offset
       sd     : signal offset in sig[1]
     ----------------------------------------------------------------------- */
  inline void analyze (double *sig[],const long len,const long step,
		       const int fi, const long offset, const long fd,
		       const long sd) const;

  /* -----------------------------------------------------------------------
     analysis filtering  - single level
       sig[0] : un-filtered signal (input)
       sig[1] : low / high pass filtered signal (output)
       len    : length of signals
       step   : sample distance in signals
       offset : index of first sample signals
       cpy    : copy high-pass filtered signal from sig[1] to sig[0]?
     ----------------------------------------------------------------------- */
  virtual inline void single (double *sig[], const long len, 
			      const long step=1, const long offset=0,
			      const unsigned int cpy=0) const;

  /* -----------------------------------------------------------------------
     analysis filtering - multi level
       sig[0] : signal (input/output)
       sig[1] : working space (same size as sig[0]) (changed)
       size   : length of signal sig[0] (in each dimension)
                i.e. sig[0] is a vector, square, cube or hypercube
       dims   : number of dimensions of signal sig[0]
       depth  : number of steps per dim. 
                (depth<=0 : max possible number + depth)
       return : pointer to result (current sig[0])
     ----------------------------------------------------------------------- */
  virtual double* multi (double *sig[], const long size, 
			 const long dims=1, long depth=0) const;

};


/* -------------------------------------------------------------------------
   analyze filter
   ------------------------------------------------------------------------- */
inline void dwtAnalyzer::analyze (double *sig[], const long len, 
				  const long step, const int fi,
				  const long offset, const long fd, 
				  const long sd) const 
{
  double *x = &(sig[0][offset]);
  double *y = &(sig[1][offset+sd]);

  for (long yi=0; yi<len/2; yi++) {
    for (long ci=0; ci<n[fi]; ci++) {
      y[yi*step] += x[index(2*yi+ci-fd,len)*step] * c[fi][ci];
    }
  }
}


/* -------------------------------------------------------------------------
   analysis filtering  - single level
   ------------------------------------------------------------------------- */
inline void dwtAnalyzer::single (double *sig[], const long len, 
			         const long step, const long offset, 
				 const unsigned int cpy) const 
{ 
  // offsets
  const long hd = signalOffset(len,1)*step; // offset of high-pass-signal

  // init sig[1] (low/high)
  init (&(sig[1][offset]),0,len,step);

  // filter-steps
  apply (sig,len,step,1,offset);   // high-pass-filter
  apply (sig,len,step,0,offset);   // low-pass-filter

  // copy high-pass-filtered signal from sig[1] to sig[0]
  if (cpy) {
    copy (&(sig[1][offset+hd]),&(sig[0][offset+hd]),len/2,step); 
  }
}


/* -------------------------------------------------------------------------
   analysis filtering - multi level
   ------------------------------------------------------------------------- */
double* dwtAnalyzer::multi (double *sig[], const long size, 
			    const long dims, long depth) const 
{
  // get depth
  depth = getDepth (size,depth);

  const long levels = depth*dims;       // number of tiling-levels
  double *result = sig[odd(levels)];    // result contains result
  long len[dims], step[dims];           // nr of elements / sample distances

  // init len / step
  initSize (len, step, dims, size, size);

  // tiling all dimensions depth times (bisect length of all dims.)
  for (long lev=0; lev<levels; depth--) {
    for (long d=0; d<dims; d++, lev++) {
      tile (sig, sig[0]==result, dims-1, d, 0, len, step);
      toggleSignals (sig,0,1);
      len[d] /= 2;
    }
  }

  // result is always sig[0] (because of toggling sig[0] and sig[1])
  return (result);
}



/* =========================================================================
   Discrete Wavelet Transformation - Filter
   ( synthesis filter )
   ========================================================================= */
class dwtSynthesizer : public dwtFilter {

public:

  /* -----------------------------------------------------------------------
     construct / destruct
       low       : low-pass-filter-coefficients (not the reversed version!)
       lowOrder  : number of coefficients in low
       high      : high-pass-filter-coefficients (not the reversed version!)
       highOrder : number of coefficients in high
     ----------------------------------------------------------------------- */
  dwtSynthesizer (void) {}
  dwtSynthesizer (const class dwtSynthesizer &flt) { set (flt); }
  dwtSynthesizer (const double *low,  const long lowOrder,
		  const double *high, const long highOrder) { 
    set (low,lowOrder, high,highOrder); }
  dwtSynthesizer (const double *low, const double *high, const long size[2]) {
    set (low, size[0], high, size[1]); }

  /* -----------------------------------------------------------------------
     set
       low       : low-pass-filter-coefficients (not the reversed version!)
       lowOrder  : number of coefficients in low
       high      : high-pass-filter-coefficients (not the reversed version!)
       highOrder : number of coefficients in high
     ----------------------------------------------------------------------- */
  virtual void set (const double *low,  const long lowOrder,
		    const double *high, const long highOrder) {
    dwtFilter::set (high,highOrder,low,lowOrder);     // low <-> high
    revert (c[0],n[0]); revert (c[1],n[1]); }
  virtual void set (const double *low, const double *high, const long size[2]){
    set (low, size[0], high, size[1]); }
  virtual void set (const class dwtSynthesizer &flt) { 
    dwtFilter::set (flt.c[0], flt.n[0], flt.c[1], flt.n[1]); }

  /* -----------------------------------------------------------------------
     interface to the filter method
       sig[0] : un-filtered signal (output)
       sig[1] : filtered signal (input)
       len    : length of sig[0], length of sig[1] = len/2
       step   : sample distance in signals
       fi     : filter-selection ( low-pass = 0 / high-pass-filter = 1 )
       offset : index of first sample in signals
     ----------------------------------------------------------------------- */
  virtual inline void apply (double *sig[], const long len, const long step, 
			     const int fi, const long offset=0) const {
    synthesize (sig, len, step, fi, offset, 
		filterOffset(), signalOffset(len,fi)*step); }

  /* -----------------------------------------------------------------------
     synthesize filter
       sig[0] : un-filtered signal (output)
       sig[1] : filtered signal (input)
       len    : length of sig[0], length of sig[1] = len/2
       step   : sample distance in signals
       fi     : filter-selection ( low-pass = 0 / high-pass-filter = 1 )
       offset : index of first sample in signals
       fd     : filter offset
       sd     : signal offset in sig[1]
     ----------------------------------------------------------------------- */
  inline void synthesize (double *sig[],const long len,const long step,
			  const int fi, const long offset, const long fd,
			  const long sd) const;

  /* -----------------------------------------------------------------------
     synthesis filtering  - single level
       sig[0] : un-filtered signal (output)
       sig[1] : low / high pass filtered signal (input)
       len    : length of signals
       step   : sample distance in signals
       offset : index of first sample in signals
       cpy    : copy high-pass filtered sig. from sig[0] to sig[1]?
     ----------------------------------------------------------------------- */
  virtual inline void single (double *sig[], const long len, 
			      const long step=1, const long offset=0,
			      const unsigned int cpy=0) const;

  /* -----------------------------------------------------------------------
     synthesis filtering - multi level
       sig[0] : signal (input/output)
       sig[1] : working space (same size as sig[0]) (changed)
       size   : length of signal sig[0] (in each dimension)
                i.e. sig[0] is a vector, square, cube or hypercube
       dims   : number of dimensions of signal sig[0]
       depth  : number of steps per dim. 
                (depth<=0 : max possible number + depth)
       return : pointer to result (current sig[0])
     ----------------------------------------------------------------------- */
  virtual double* multi (double *sig[], long size, 
			 const long dims=1, long depth=0) const;

};


/* -------------------------------------------------------------------------
   apply filter
   ------------------------------------------------------------------------- */
inline void dwtSynthesizer::synthesize (double *sig[], const long len, 
					const long step, const int fi, 
					const long offset, const long fd,
					const long sd) const
{
  double *x = &(sig[0][offset]);
  double *y = &(sig[1][offset+sd]);

  for (long yi=0; yi<len/2; yi++) {
    for (long ci=0; ci<n[fi]; ci++) {
      x[index(2*yi-ci+fd,len)*step] += y[yi*step] * c[fi][ci];
    }
  }
}


/* -------------------------------------------------------------------------
   synthesis filtering  - single level
   ------------------------------------------------------------------------- */
inline void dwtSynthesizer::single (double *sig[], const long len, 
				    const long step, const long offset, 
				    const unsigned int cpy) const 
{
  // offsets
  const long hd = signalOffset(len,1)*step; // offset of high-pass-signal

  // copy high-pass-filtered signal from x to y
  if (cpy) {
    copy (&(sig[0][offset+hd]),&(sig[1][offset+hd]),len/2,step); 
  }
  
  // init x
  init (&(sig[0][offset]),0,len,step);

  // reverse filter-steps
  apply (sig,len,step,0,offset);   // reverse low-pass-filter
  apply (sig,len,step,1,offset);   // reverse high-pass-filter
}


/* -------------------------------------------------------------------------
   synthesis filtering - multi level
   ------------------------------------------------------------------------- */
double* dwtSynthesizer::multi (double *sig[], const long size, 
			       const long dims, long depth) const
{
  // get depth
  depth = getDepth (size,depth);

  const long levels = depth*dims;      // number of tiling-levels
  double *input = sig[0];              // original input
  long len[dims], step[dims];          // nr of elements / sample distances

  // init len / step
  initSize (len, step, dims, size>>depth, size);
    
  // un-tiling all dimensions depth times (double length of all dims.)
  for (long lev=0; lev<levels; depth--) {
    for (long d=dims-1; d>=0; d--, lev++) {
      len[d] *= 2;
      toggleSignals (sig,0,1);
      tile (sig, sig[1]!=input, dims-1, d, 0, len, step);
    }
  }

  // result is always sig[0] (because of toggling sig[0] and sig[1])
  return (sig[0]);
}


/* =========================================================================
   Discrete Wavelet Transformation - Filter
   ( synthesis filter - get weights for reverse-high-pass filters )
   ========================================================================= */
class dwtNoise : public dwtSynthesizer {

protected:

  /* -----------------------------------------------------------------------
     attributes
     ----------------------------------------------------------------------- */
  double delta; // values >= delta are weighted by 1.0 
  long shifts;
  long skip;

public:

  /* -----------------------------------------------------------------------
     construct / destruct
       low       : low-pass-filter-coefficients (not the reversed version!)
       lowOrder  : number of coefficients in low
       high      : high-pass-filter-coefficients (not the reversed version!)
       highOrder : number of coefficients in high
     ----------------------------------------------------------------------- */
  dwtNoise (void) {}
  dwtNoise (const class dwtNoise &flt) { 
    set (flt); }
  dwtNoise (const double *low,  const long lowOrder,
	    const double *high, const long highOrder) { 
    set (low,lowOrder, high,highOrder); }
  dwtNoise (const double *low, const double *high, const long size[2]) {
    set (low, size[0], high, size[1]); }

  /* -----------------------------------------------------------------------
     set parameters
     ----------------------------------------------------------------------- */
  void setParams (const double level1, 
		  const long shiftLevels, const long skipLevels=0) { 
    delta = level1; shifts = shiftLevels; skip = skipLevels; }

  /* -----------------------------------------------------------------------
     interface to the filter method
       sig[0] : un-filtered signal (output)
       sig[1] : filtered signal (input)
       len    : length of sig[0], length of sig[1] = len/2
       step   : sample distance in signals
       fi     : filter-selection ( low-pass = 0 / high-pass-filter = 1 )
       offset : index of first sample in signals
     ----------------------------------------------------------------------- */
  virtual inline void apply (double *sig[], const long len, const long step, 
			     const int fi, const long offset=0) const {
    weighted (sig, len, step, fi, offset, 
	      filterOffset(), signalOffset(len,fi)*step); }

  /* -----------------------------------------------------------------------
     max cumulate forward / backward
       sig[0] : un-filtered signal (output)
       sig[1] : filtered signal (input)
       len    : length of sig[0], length of sig[1], sig[2] = len/2
       step   : sample distance in signals
       fi     : filter selection (0=lowpass/1=highpass)
       offset : index of first sample in signals
       fd     : filter offset
       sd     : additional offset in sig[1]
     ----------------------------------------------------------------------- */
  inline void cumulate (double *sig[], const long len, 
			const long step, const int fi, const long offset,
			const long fd, const long sd) const;

  /* -----------------------------------------------------------------------
     weighted synthesize filter
       sig[0] : un-filtered signal (output)
       sig[1] : filtered signal (input)
       sig[2] : weights for reverse high-pass
       len    : length of sig[0], length of sig[1], sig[2] = len/2
       step   : sample distance in signals
       fi     : filter selection (0=lowpass/1=highpass)
       offset : index of first sample in signals
       fd     : filter offset
       sd     : additional offset in sig[1]
     ----------------------------------------------------------------------- */
  inline void weighted (double *sig[], const long len, 
			const long step, const int fi, const long offset,
			const long fd, const long sd) const;

  /* -----------------------------------------------------------------------
     synthesis filtering  - single level
       sig[0] : un-filtered signal (output)
       sig[1] : low / high pass filtered signal (input)
       sig[2] : weights (in mode & 2 only)
       len    : length of signals
       step   : sample distance in signals
       offset : index of first sample in signals
       mode   : mode & 1 => filter  ( 0 : low-pass / 1 : high pass )
                mode & 2 => denoise ( 0 : normal   / 2 : cumulate )
		mode & 4 => both filters (normal synthesis filtering)
     ----------------------------------------------------------------------- */
  virtual inline void single (double *sig[], const long len, 
			      const long step=1, const long offset=0,
			      const unsigned int mode=0) const;

  /* -----------------------------------------------------------------------
     noise detection filtering - multi level
       sig[0] : multi-level analysis filtered signal (input)
       sig[1] : weights (twice the size of sig[0]) (output)
       sig[2] : noise (working area, same size as sig[1]) (changed)
       size   : length of signal sig[0] (in each dimension)
                i.e. sig[0] is a vector, square, cube or hypercube
       dims   : number of dimensions of signal sig[0]
       depth  : number of steps per dim. 
                (depth<=0 => max possible number of steps + depth)
       return : pointer to result (current sig[0])
     ----------------------------------------------------------------------- */
  virtual double* multi (double *sig[], long size, 
			 const long dims=1, long depth=0) const;
  virtual double* multi2 (double *sig[], long size, 
			  const long dims=1, long depth=0) const;

protected:

  /* -----------------------------------------------------------------------
     weight-function
     ----------------------------------------------------------------------- */
  inline double convert (const double x) const { 
    const double w = 1.0 - MINMAX(ABS(x)/delta, 0.0, 1.0);
    return (w); }

  /* -----------------------------------------------------------------------
     weight-function for all values x[0..len-1]
     ----------------------------------------------------------------------- */
  inline void convert (double *x, long len, const long step=1) const {
    for (len--; len>=0; x[len*step]=convert(x[len*step]), len--); }

};


/* -------------------------------------------------------------------------
   max cumulation
   ------------------------------------------------------------------------- */
inline void dwtNoise::cumulate (double *sig[], const long len, const long step,
				const int fi, const long offset, const long fd,
				const long sd) const
{
  double *x  = &(sig[0][offset]);
  double *y  = &(sig[1][offset+sd]);

  for (long yi=0; yi<len/2; yi++) {
    const long yj = yi*step;
    for (long ci=0; ci<n[fi]; ci++) {
      const long xj = index(2*yi-ci+fd,len)*step;
      x[xj] = MAX(ABS(x[xj]), ABS(y[yj]));
    }
  }
}


/* -------------------------------------------------------------------------
   weighted high-pass filter
   ------------------------------------------------------------------------- */
inline void dwtNoise::weighted (double *sig[], const long len, const long step,
				const int fi, const long offset, const long fd,
				const long sd) const
{
  double *x = &(sig[0][offset]);
  double *y = &(sig[1][offset+sd]);
  double *w = &(sig[2][offset]);

  for (long yi=0; yi<len/2; yi++) {
    const long yj = yi*step;
    for (long ci=0; ci<n[fi]; ci++) {
      long xj = index(2*yi-ci+fd,len)*step;
      x[xj] += y[yj] * c[fi][ci] * w[xj];
    }
  }
}


/* -------------------------------------------------------------------------
   synthesis filtering  - single level
   ------------------------------------------------------------------------- */
inline void dwtNoise::single (double *sig[], const long len, 
			      const long step, const long offset,
			      const unsigned int mode) const 
{
  // init
  if (mode & 2) init (&(sig[0][offset]),0,len,step); 

  if (mode & 8) {

    // normal synthesis filtering
    synthesize (sig, len, step, mode&1, offset, filterOffset(), (len/2)*step);

  } else if (mode & 4) {
    
    // weighted synthesis filtering
    weighted (sig, len, step, mode&1, offset, filterOffset(), (len/2)*step);

  } else {

    // max cumulate
    cumulate (sig, len, step, mode&1, offset, filterOffset(), (len/2)*step);

  }
}


/* -------------------------------------------------------------------------
   synthesis filtering - multi level
   ------------------------------------------------------------------------- */
double* dwtNoise::multi (double *sig[], const long size, 
			 const long dims, long depth) const
{
  // get depth
  depth = getDepth (size,depth);

  // inititializations
  long len[dims], step[dims];              // dim-length / sample distances
  initSize (len, step, dims, size, size);  // init step and len
  const long elems  = step[dims-1]*size*2; // #elements in weight matrix
  const long levels = depth*dims;          // number of tiling-levels
  double *input  = sig[0];                 // input signal
  double *noise  = sig[1];                 // noise signal
  double *weight = sig[2];                 // weights
  
  if (delta==0.0) {

    // nothing to do
    init (weight, 0, elems, 1);
    init (noise, 0, step[dims-1]*size, 1);

  } else {

    // create weight matrix
    // shifting high-pass filtered signal using "max-filter"
    double *sig1[2] = { weight, input };
    const int sft1[1] = { 0 };                               // shift signals
    
    //init (weight, 0, elems, 1);                            // init weight
    for (long flt=1, s=0; s<shifts; s++, flt=0, sig1[1]=weight) {
      const long start = MAX(0,shifts-s-1);                  // start level
      shift(sig1,sft1,1,levels-s,start,dims,len,step,flt|2); // shift all lev.
      for (long i=0; i<dims; len[i]=size, i++);              // (re-)init len
    }  
    convert (weight, elems, 1);                              // shift -> weight
    
    // cumulating noise
    // shifting using weighted synthesis filters this time
    // highest level is directly shifted to noise
    double *sig2[3] = { noise, input, &(weight[elems/2]) }; 
    double *sig3[3] = { input, input, weight }; 
    const int sft3[2] = { 0, 2 };                            // shift signals
    
    init (noise, 0, step[dims-1]*size, 1);                   // init noise
    for (long flt=1, s=0; s<levels-shifts; s++, flt=0) {
      const long fst = (s<skip ? 8 : 0);
      tile (sig2, flt|4|fst, dims-1, 0, 0, len, step);       // shift 1st lev.
      shift (sig3,sft3,2, levels-s,1, dims,len,step, flt|6|fst); // other lev.
      for (long i=0; i<dims; len[i]=size, i++);              // (re-)init len
    }

  }

  // make noise first signal
  toggleSignals (sig,0,1);
  return (noise);
}


/* -------------------------------------------------------------------------
   synthesis filtering - multi level
   ------------------------------------------------------------------------- */
double* dwtNoise::multi2 (double *sig[], const long size, 
			  const long dims, long depth) const
{
  // get depth
  depth = getDepth (size,depth);

  // inititializations
  long len[dims], step[dims];              // dim-length / sample distances
  initSize (len, step, dims, size, size);  // init step and len
  const long elems  = step[dims-1]*size;   // #elements in weight matrix
  const long levels = depth*dims;          // number of tiling-levels
  double *input  = sig[0];                 // input signal
  double *noise  = sig[1];                 // noise signal
  double *weight = sig[2];                 // weights
  
  if (delta==0.0) {

    // nothing to do
    init (weight, 0, elems, 1);
    init (noise, 0, step[dims-1]*size, 1);

  } else {

    // create weight matrix
    // shifting high-pass filtered signal using "max-filter"
    double *sig1[2] = { weight, input };
    const int sft1[1] = { 0 };                               // shift signals
    
    //init (weight, 0, elems, 1);                            // init weight
    for (long flt=1, s=0; s<shifts; s++, flt=0, sig1[1]=weight) {
      const long start = MAX(1,shifts-s);                    // start level
      shift(sig1,sft1,1,levels-s,start,dims,len,step,flt|2); // shift all lev.
      for (long i=0; i<dims; len[i]=size, i++);              // (re-)init len
    }  
    convert (weight, elems, 1);                              // shift -> weight
    
    // cumulating noise
    // shifting using weighted synthesis filters this time
    // highest level is directly shifted to noise
    double *sig2[3] = { noise, input, weight }; 
    double *sig3[3] = { input, input, weight }; 
    const int sft3[1] = { 0 };                               // shift signals
    
    init (noise, 0, step[dims-1]*size, 1);                   // init noise
    for (long flt=1, s=0; s<levels-shifts; s++, flt=0) {
      const long fst = (s<skip ? 8 : 0);
      tile (sig2, flt|4|fst, dims-1, 0, 0, len, step);       // shift 1st level
      shift (sig3,sft3,1, levels-s,1, dims,len,step, flt|6|fst); // other lev.
      for (long i=0; i<dims; len[i]=size, i++);              // (re-)init len
    }

  }

  // make noise first signal
  toggleSignals (sig,0,1);
  return (noise);
}


/* =========================================================================
   Selective Blur
   ========================================================================= */
class dwtBlur : public dwtSynthesizer {

protected:

  /* -----------------------------------------------------------------------
     attributes
     ----------------------------------------------------------------------- */
  double delta;    // max delta
  long radius;     // blur radius

public:

  /* -----------------------------------------------------------------------
     construct / destruct
       low       : low-pass-filter-coefficients (not the reversed version!)
       lowOrder  : number of coefficients in low
       high      : high-pass-filter-coefficients (not the reversed version!)
       highOrder : number of coefficients in high
     ----------------------------------------------------------------------- */
  dwtBlur (void) {}
  dwtBlur (const class dwtBlur &flt) { set(flt); }
  dwtBlur (const double *low,  const long lowOrder,
	   const double *high, const long highOrder) { 
    set (low,lowOrder, high,highOrder); }
  dwtBlur (const double *low, const double *high, const long size[2]) {
    set (low, size[0], high, size[1]); }

  /* -----------------------------------------------------------------------
     set parameters
     ----------------------------------------------------------------------- */
  void setParams (const double maxDelta, const long blurRadius) { 
    delta = maxDelta; radius = blurRadius; }

  /* -----------------------------------------------------------------------
     synthesis filtering - multi level
       sig[0] : signal (input/output)
       sig[1] : working space (same size as sig[0]) (changed)
       sig[2] : weights (input)
       size   : length of signal sig[0] (in each dimension)
                i.e. sig[0] is a vector, square, cube or hypercube
       dims   : number of dimensions of signal sig[0]
       depth  : number of steps per dim. 
                (depth<=0 : max possible number + depth)
       return : pointer to result (current sig[0])
     ----------------------------------------------------------------------- */
  virtual double* multi (double *sig[], long size, 
			 const long dims=1, long depth=0) const;

  /* -----------------------------------------------------------------------
     denoise
       signal : signal to denoise
       noise  : signal = signal-noise
       len    : number of elements
       step   : sample distance
     ----------------------------------------------------------------------- */
  inline void denoise (double *signal, double *noise, 
		       long len, const long step=1) const {
    for (len--; len>=0; signal[len]-=noise[len], len--); }

  /* -----------------------------------------------------------------------
     selective blur
       input  : input signal
       output : blurred signal (output)
       weight : weights (0:blur ... 1:don't blur)
       dims   : number of dimensions
       len    : length per dimension
       step   : sample distance per dimension
     ----------------------------------------------------------------------- */
  void blur (double *input, double *output, double *weight,
	     const long dims, const long *len, const long *step) const {
    long elem[dims]; for (long d=0; d<dims; elem[d]=0, d++);
    blur (input, output, weight, dims, dims-1, 0, elem, len, step, radius); }

protected:

  /* -----------------------------------------------------------------------
     selective blur
       input  : input signal
       output : blurred signal
       weight : weights (0:blur ... 1:don't blur)
       dims   : number of dimensions
       dim    : last dimension
       offset : index of first element
       elem   : coordinates of current element
       len    : length per dimension
       step   : sample distance per dimension
       rad    : radius
     ----------------------------------------------------------------------- */
  void blur (double *input, double *output, double *weight,
	     const long dims, const long dim, long offset,
	     long *elem, const long *len, const long *step,
	     const long rad) const;

  /* -----------------------------------------------------------------------
     blur area
       dims   : number of dimensions
       elem   : current coordinates
       len    : length per dimension
       step   : sample distance per dimension
       rad    : radius
       offset : first element (input/output)
       area   : length of area per dimension (output)
     ----------------------------------------------------------------------- */
  inline void blurArea (const long dims, const long *elem, const long *len,
			const long *step, const long rad, long &offset, 
			long *area) const;

  /* -----------------------------------------------------------------------
     weighted sum
       sig    : signal
       x      : value to compare with
       dim    : last dimension
       offset : first element
       len    : length for all dimensions
       step   : sample distance per dimension
       w      : cumulated weights (output)
       v      : weighted sum (output)
     ----------------------------------------------------------------------- */
  void sum (double *sig, const double x, const long dim, long offset, 
	    const long *len, const long *step, double &w, double &v) const;

  /* -----------------------------------------------------------------------
     weight-function 
     ----------------------------------------------------------------------- */
  inline double convert (const double x) const { 
    const double w = 1.0 - MINMAX(ABS(x)/delta, 0.0, 1.0);
    return (w); }

};


/* -------------------------------------------------------------------------
   synthesis filtering - multi level
   ------------------------------------------------------------------------- */
double* dwtBlur::multi (double *sig[], const long size, 
			const long dims, long depth) const
{
  // get depth
  depth = getDepth (size,depth);

  const long levels = depth*dims;         // number of tiling-levels
  double *input = sig[0];                 // original input
  long len[dims], step[dims], elem[dims]; // nr of elements / sample distances

  // init len / step
  initSize (len, step, dims, size>>depth, size);
    
  // un-tiling all dimensions depth times (double length of all dims.)
  toggleSignals (sig,0,1);
  for (long rad=radius, lev=0; lev<levels; depth--, rad/=2) {
    for (long d=dims-1; d>=0; d--, lev++) {
      len[d] *= 2;
      tile (sig, sig[1]!=input, dims-1, d, 0, len, step);
      for (long i=0; i<dims; elem[i]=0, i++);
      const long offset = shiftOffset(d, dims, len, step);
      blur (sig[0],sig[1],&(sig[2][offset]), dims,dims-1, 0,elem,len,step,rad);
    }
  }

  // result is always sig[0] (because of toggling sig[0] and sig[1])
  toggleSignals (sig,0,1);
  return (sig[0]);
}


/* -------------------------------------------------------------------------
   selective blur
   ------------------------------------------------------------------------- */
void dwtBlur::blur (double *input, double *output, double *weight,
		    const long dims, const long dim, long offset, 
		    long *elem, const long *len, const long *step,
		    const long rad) const 
{
  if (dim<0) {

    if (delta<=0.0) {

      // just copy
      output[offset] = input[offset];

    } else {

      // get weighted sum in area
      double w = 0, v = 0;
      long offsetLocal = offset, lenLocal[dims];
      blurArea (dims, elem, len, step, rad, offsetLocal, lenLocal);
      sum (input, input[offset], dims-1, offsetLocal, lenLocal, step, w, v);
      
      // blur
      if (w==0.0) {
	output[offset] = input[offset];
      } else {
	double wgt = weight[offset] * weight[offset];  // ( squared weight )
	output[offset] = (wgt * input[offset]) + ((1.0-wgt) * v/w);
      }

    }

  } else {

    // recursion
    for (elem[dim]=0; elem[dim]<len[dim]; offset+=step[dim], elem[dim]++) {
      blur (input,output,weight, dims,dim-1, offset, elem,len,step, rad);
    }

  }
}


/* -------------------------------------------------------------------------
   blur area
   ------------------------------------------------------------------------- */
inline void dwtBlur::blurArea (const long dims, const long *elem, 
			       const long *len, const long *step, 
			       const long rad, long &offset, long *area) const 
{
  for (long d=0; d<dims; d++) {
    const long start = -MIN(rad,elem[d]);
    offset += start * step[d];
    area[d] = MIN(rad,len[d]-elem[d]-1)-start;
  }
}


/* -------------------------------------------------------------------------
   weighted sum
   ------------------------------------------------------------------------- */
void dwtBlur::sum (double *sig, const double x, const long dim, 
		   long offset, const long *len, const long *step,
		   double &w, double &v) const
{
  if (dim<0) {

    // sum
    const double c = convert (x-sig[offset]); // current weight
    w += c;                                   // cumulate weights
    v += sig[offset] * c;                     // weighted sum

  } else {
    
    // recursion
    for (long i=0; i<len[dim]; offset+=step[dim], i++) {
      sum (sig, x, dim-1, offset, len, step, w, v);
    }
  }

}

#endif // ifndef _WAVELET_CPP


